import 'package:plexus/expose.dart';

class Lang extends UnmodifiableMapView<String, dynamic> {
  const Lang(this.locale, Map<String, dynamic> localizations) : super(localizations);

  final Locale locale;

  bool isDirectlySupported(Locale locale) => Lang.supportedLocales.contains(locale);
  bool isSupported(Locale locale) => Lang.supportedLanguages.contains(locale.language);

  dynamic operator [](dynamic key) => super[key] ?? key;

  static const supportedLanguages = Locale.languages;
  static const supportedCountries = Locale.countries;
  static const supportedLocales = Locale.locales;

  static const ENGLISH__COMMON = {
    'greeting': 'Welcome',
    'enthusiastic_greeting': 'Welcome!'
  };

  static const ENGLISH__CANADA = {
    ...ENGLISH__COMMON,
  };
}

class Locale {
  const Locale(this.language, this.country);

  final String language;
  final String country;

  static const COUNTRY_CANADA = 'CA';
  static const COUNTRY_UNITED_KINGDOM = 'UK';
  static const COUNTRY_UNITED_STATES = 'US';

  static const LANGUAGE_ENGLISH = 'en';

  static const languages = {
    LANGUAGE_ENGLISH
  };

  static const countries = {
    COUNTRY_CANADA,
    COUNTRY_UNITED_KINGDOM,
    COUNTRY_UNITED_STATES,
  };

  static const locales = {
    Locale(LANGUAGE_ENGLISH, COUNTRY_CANADA),
    Locale(LANGUAGE_ENGLISH, COUNTRY_UNITED_KINGDOM),
    Locale(LANGUAGE_ENGLISH, COUNTRY_UNITED_STATES)
  };
}
