import 'dart:async';
import 'dart:collection';

import 'package:plexus/plexus.dart';
import 'package:tuple/tuple.dart';

class PatientStreamQueue<E> {
  PatientStreamQueue.internal(this.controller, this._listeners) : _queue = Queue(), $=IndexingOperatorWrapper((k) => _listeners[k], (k,v) => _listeners[k] = v, (k) => _listeners.remove(k)), broadcastStream = controller.stream.asBroadcastStream(), sink = controller.sink {
    controller.onListen = _onListen;
    if (controller.hasListener) _onListen();
  }

  PatientStreamQueue() : this.internal(StreamController(), {});

  final StreamController<E> controller;
  Stream<E> get stream => controller.stream;
  final Stream<E> broadcastStream;
  final Sink<E> sink;

  final Queue<dynamic> _queue;

  bool listened = false;

  Map<String, Tuple2<FutureOrValueCallback<E, void>, StreamSubscription<E>>> _listeners;

  /// Map of keys to queue listeners.
  final IndexingOperatorWrapper<String, Tuple2<FutureOrValueCallback<E, void>, StreamSubscription<E>>> $;

  bool get isClosed => controller.isClosed;

  void add(E value) => listened ? controller.add(value) : _queue.add(value);
  void addError(Object error, [@nullable StackTrace stackTrace]) => listened ? controller.addError(error, stackTrace) : _queue.add(Tuple2(error, stackTrace));

  Future<void> dispose() async {
    await controller.close();
  }

  void _onListen() {
    listened = true;
    
    while (_queue.isNotEmpty) {
      final current = _queue.removeFirst();
      if (current is Tuple2<Object, StackTrace>) controller.addError(current.item1, current.item2);
      else controller.add(current);
    }
  }
}

// class BufferedQueueElement<E> {
//   const BufferedQueueElement(this.value);

//   final E value;
// }

// /// Time=priority based queue.
// ///
// /// Keeps a cyclic buffer of elements, and voids older elements when it fills up.
// ///
// /// The structure is efficient for any queue or stack usage.
// class BufferedQueue<E> {

//   /// Create an empty queue.
//   ///
//   /// If [capacity] is given, the queue will hold that many elements. If
//   /// [capacity] is null or is less than [MINIMUM_CAPACITY], the queue will hold [MINIMUM_CAPACITY] elements instead.
//   BufferedQueue([int capacity = MINIMUM_CAPACITY, this.keepFirst = true]) : _queue = List((capacity < MINIMUM_CAPACITY) ? MINIMUM_CAPACITY /*: capacity > MAXIMUM_CAPACITY ? MAXIMUM_CAPACITY*/ : capacity);

//   /// Create an queue with [initial] items.
//   ///
//   /// An error will be thrown if [_queue.length] is not a power of two.
//   BufferedQueue.startsWith(List<E?> queue, [this.keepFirst = true])
//       : assert(queue.length >= MINIMUM_CAPACITY, 'Queue must be at least as large as `[MINIMUM_CAPACITY] == $MINIMUM_CAPACITY`') /*, assert(queue.length > MAXIMUM_CAPACITY, 'Queue must be no larger than `[MAXIMUM_CAPACITY] == $MAXIMUM_CAPACITY`) ? MAXIMUM_CAPACITY*/,
//         _queue = queue.map<BufferedQueueElement<E>?>((e) => BufferedQueueElement<E?>(e)).toList();

//   final List<BufferedQueueElement<E>?> _queue;

//   /// Should this queue always have an element
//   final bool keepFirst;
  
//   /// true if the queue is empty
//   bool get isEmpty => _queue.first != null;

//   /// true if the queue is not empty
//   bool get isNotEmpty => _queue.first == null;

//   /// The length of the queue
//   int get length => _queue.length;

//   /// Returns the number of slots in the queue that are used (not-null)
//   int get lengthUsed => _queue.where((e) => e != null).length;

//   int get efficientLengthUsed => hasMultipleUsed ? lengthUsed : hasAnyUsed ? 1 : 0;

//   /// Returns the number of slots in the queue that are unused (null)
//   int get lengthUnused => _queue.where((e) => e == null).length;

//   bool get hasAnyUsed => _queue[0] != null;

//   bool get hasMultipleUsed => _queue[0] != null && _queue[1] != null;

//   final StreamController<E> _controller = StreamController();

//   /// The first elemeent in the queue. This does *not* remove the first element when used. To do so, use `pop()` instead.
//   E? get first {
//     return _queue.first?.value;
//   }

//   /// Removes and returns the first element in the queue, disrespecting [keepFirst].
//   E? get pop => popFunc(respectKeepFirst: false);

//   /// Removes and returns the first element in the queue.
//   E? popFunc({bool respectKeepFirst = true}) {
//     if (keepFirst && respectKeepFirst && !hasMultipleUsed) {
//       return first;
//     }
    
//     final ret = first;
//     removeFirst();
//     return ret;
//   }

//   /// Adds [value] to the end of the queue
//   void add(E value) {
//     final couldMakeRoom = makeRoom(1, respectKeepFirst: false);
//     final len = efficientLengthUsed;
//     if (!couldMakeRoom) {
//       throw IndexError(len, toList(), 'NotEnoughSpaceError', 'Could not make room for new value `$value` of type `$E`.');
//     }
//     _queue[len] = BufferedQueueElement(value);
//     _controller.add(value);
//   }

//   /// Removes the first element in the queue.
//   void removeFirst({bool respectKeepFirst = true}) => shift(1, respectKeepFirst: respectKeepFirst);

//   /// Clears the queue
//   void clear({bool respectKeepFirst = true}) {
//     if (!hasMultipleUsed) {
//       return;
//     }

//     final j = lengthUsed - 1 - ((keepFirst && respectKeepFirst) ? 1 : 0);

//     for (var i = 0; i < j; i++) {
//       _queue[i] = null;
//     }

//     return;
//   }

//   /// Loops through the queue, performing [action] on each element, then removing them all at once.
//   void forEach(void Function(E element) action, {bool respectKeepFirst = true}) {
//     final j = efficientLengthUsed;

//     for (var i = 0; i < j; i++) {
//       action(_queue[i]!.value);
//     }
//     clear(respectKeepFirst: respectKeepFirst);
//     //OR
//     /// Loops through the queue, performing [action] on and then removing each element.
//     //while (first != null) {
//     //  action(first);
//     //  shift(1);
//     //}
//     //OR
//     /// Loops through the queue, popping and performing [action] on each element.
//     //while (first != null) {
//     //  action(pop);
//     //}
//   }

//   /// Removes and returns a [List<E>] containing the first [count] elements in the queue.
//   List<E?> take(int count, {bool respectKeepFirst = true}) {
//     if (count > length) {
//       throw IndexError(count, toList());
//     } else if (count == length) {
//       final l = toList();
//       clear(respectKeepFirst: respectKeepFirst);
//       return l;
//     }

//     //ignore: argument_type_not_assignable
//     final taken = _queue.take(count).map<E?>(_unpackageElement).toList(growable: false);
//     shift(count);
//     return taken;
//   }

//   /// Returns a shallow-copy of the queue.
//   //ignore: argument_type_not_assignable
//   List<E?> toList({bool growable = false}) => _queue.map<E?>(_unpackageElement).toList(growable: growable);

//   /// Returns an fixed-size shallow-copy of the queue.
//   //ignore: argument_type_not_assignable
//   Set<E?> toSet() => _queue.map<E?>(_unpackageElement).toSet();

//   /// Shifts the queue towards the start by [by], voiding elements.
//   /// 
//   /// Returns false if shifting was entirely cancelled. This will happen if
//   /// shifting would remove the first element.
//   bool shift(int by, {bool respectKeepFirst = true}) {
//     if (by < 1) {
//       // TODO(mpfaff): Log warning that because `by < 1`, nothing will happen and this was a waste of time.
//       return true;
//     }

//     final j = efficientLengthUsed;

//     if (by >= j) {
//       // TODO(mpfaff): Log warning that because `by >= efficientLengthUsed`, `clear()` will be used instead for efficiency.
//       clear(respectKeepFirst: respectKeepFirst);
//       return true;
//     }

//     if (keepFirst && respectKeepFirst) {
//       if (by >= j - 1) {
//         // TODO(mpfaff): Log warning that because `by >= j - 1` and `contextKeepFirst` is true, `by` will be set to `j - 2`.
//         //by = j - 2;
//         return false;
//       }

//       for (var i = 0; i < j - 1 && i + by < _queue.length; i++) {
//         if (i + by < _queue.length) {
//           _queue[i] = _queue[i + by];
//           _queue[i + by] = null;
//         } else if (i != j - 1) {
//           _queue[i] = null;
//         }
//       }
//     } else {
//       for (var i = 0; i < j && i + by < _queue.length; i++) {
//         if (i + by < _queue.length) {
//           _queue[i] = _queue[i + by];
//           _queue[i + by] = null;
//         } else {
//           _queue[i] = null;
//         }
//       }
//     }

//     return true;
//   }

//   /// Makes room in the queue for [slots] slots
//   bool makeRoom(int slots, {bool respectKeepFirst: true}) {
//     if (slots > length) {
//       return false;
//     }
    
//     final l = lengthUnused;

//     if (l >= slots) {
//       return true;
//     }

//     return shift(slots - l, respectKeepFirst: respectKeepFirst);
//   }

//   /// Listen for new elements in the queue.
//   void listen(ValueCallback<E> callback) {
//     _controller.stream.listen(callback);
//   }

//   E? _unpackageElement(BufferedQueueElement<E>? container) => container?.value;
//   BufferedQueueElement<E> _packageElement(E value) => BufferedQueueElement<E>(value);

//   static const int MINIMUM_CAPACITY = 4;
//   //9223372036854775806 == int64MaxValue - 1
//   //9,223,372,040,000,000,000
//   //9,223,372,036,854,775,806
//   //static const int MAXIMUM_CAPACITY = 9223372036854775806;

//   // Internal helper functions.

//   /// Whether [number] is a power of two.
//   ///
//   /// Only works for positive numbers.
//   static bool _isPowerOf2(int number) => (number & (number - 1)) == 0;

//   /// Rounds [number] up to the nearest power of 2.
//   ///
//   /// If [number] is a power of 2 already, it is returned.
//   ///
//   /// ~~Only works for positive numbers.~~ If [number] is below 2, regardless of being a power of 2, 2 is returned.
//   static int _nextPowerOf2(int number) {
//     if (number < 2) {
//       return 2;
//     }
//     // assert(number > 0, 'Number must be greater than 0'); // why?
//     number = (number << 1) - 1;
//     for (int nextNumber = number & (number - 1); nextNumber != 0; number = nextNumber, nextNumber = number & (number - 1)) {
//       // final int nextNumber = number & (number - 1);
//       // if (nextNumber == 0) {
//       //   return number;
//       // }
//       // number = nextNumber;
//     }

//     return number;
//   }
// }
