import 'dart:async';
import 'dart:io';

import 'package:plexus/plexus.dart';

/// The user's home directory
Directory get userHome => Directory(Platform.environment['HOME'] ?? Platform.environment['USERPROFILE'] ?? '/sdcard');

/// Can be used to bypass a parse function.
T loopback<T>(T t) => t;

/// Prints [value] using the native [print] method, optionally appending and prepending newlines to [value].
void printSpaced(dynamic value, {bool leadingNewline = false, bool trailingNewline = true}) => print('${leadingNewline ? '\n' : ''}$value${trailingNewline ? '\n' : ''}');

/// A [RegExp] pattern for all horizontal whitespace characters,
final whitespacePattern = RegExp(r'\p{Blank}');

class Nullable {
  const Nullable._();
}

const Nullable nullable = Nullable._();

typedef void ValueCallback<T>(T value);
typedef bool ValueTest<T>(T value);

typedef FutureOr<R> FutureOrValueCallback<V, R>(V value);
typedef FutureOr<R> FutureOrCallback<R>();

typedef void VoidCallback();
typedef Future<void> AsyncVoidCallback();

typedef E GetListEntryCallback<E>(int index);
typedef void SetListEntryCallback<E>(int index, E value);
typedef void RemoveListEntryCallback(int index);

typedef V GetEntryCallback<K, V>(K key);
typedef void SetEntryCallback<K, V>(K key, V value);
typedef void RemoveEntryCallback<K>(K key);


void voidEventHandler(Event _) {}
Future<void> asyncVoidEventHandler(Event _) async {}

void voidHandler() {}
Future<void> asyncVoidHandler() async {}


abstract class OperatorWrapper {

}

class IndexingOperatorWrapper<K, V> implements OperatorWrapper {
  const IndexingOperatorWrapper(this.getCallback, this.setCallback, this.rmCallback);

  final GetEntryCallback<K, V> getCallback;
  final SetEntryCallback<K, V> setCallback;
  final RemoveEntryCallback<K> rmCallback;

  /// Returns the value of `key`.
  V operator [](K key) => getCallback(key);

  /// Sets the value of `key` to `value`.
  void operator []=(K key, V value) => setCallback(key, value);

  /// Removes `key` and its value
  void operator |(K key) => rmCallback(key);
}

class ListOperatorWrapper<E> extends IndexingOperatorWrapper<int, E> {
  ListOperatorWrapper(GetEntryCallback<int, E> getCallback, SetEntryCallback<int, E> setCallback, RemoveEntryCallback<int> rmCallback) : super(getCallback, setCallback, rmCallback);
}
